#! coding: utf-8
from selectel_storage.base import SelectelApi, mongoengine
from selectel_storage.exceptions import SelectelStorageExeption


class SelectelStorage(object):

    def __init__(self, app=None):
        self.mongoengine = mongoengine
        if app:
            self.init_app(app)

    def init_app(self, app):
        app.config.setdefault('SELECTEL_STORAGE', {})

        if not app.config.get('SELECTEL_STORAGE'):
            raise SelectelStorageExeption('Bad config: No SELECTEL_STORAGE options')

        self.connection = SelectelApi(
            user=app.config['SELECTEL_STORAGE']['USER'],
            password=app.config['SELECTEL_STORAGE']['PASSWORD'],
            container=app.config['SELECTEL_STORAGE']['CONTAINER']
        )

        app.extensions = getattr(app, 'extensions', {})
        app.extensions['selectel_storage'] = self



#! coging: utf-8
from __future__ import unicode_literals

import gzip
from StringIO import StringIO

import magic
import time
from requests.exceptions import HTTPError, ReadTimeout
from selectel_storage.base.storage import Storage
from functools import wraps


def retry(ExceptionToCheck, tries=4, delay=3, backoff=2, logger=None):
    def deco_retry(f):
        @wraps(f)
        def f_retry(*args, **kwargs):
            mtries, mdelay = tries, delay
            while mtries > 1:
                try:
                    return f(*args, **kwargs)
                except ExceptionToCheck:
                    time.sleep(mdelay)
                    mtries -= 1
                    mdelay *= backoff
            return f(*args, **kwargs)
        return f_retry
    return deco_retry


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class SelectelApi(object):
    __metaclass__ = Singleton

    def __init__(self, user=None, password=None, container=None):
        self.user = user
        self.password = password
        self.container = container
        self.connection = Storage(self.user, self.password)

    def add(self, path, file_obj):
        self._make_call('put', self.container, path, file_obj)

    def get(self, path):
        return self._make_call('get_stream', self.container, path)

    def delete(self, path):
        self._make_call('remove', self.container, path)

    @retry((ReadTimeout, HTTPError), delay=1.5)
    def _make_call(self, method, *args):
        try:
            return getattr(self.connection, method)(*args)
        except HTTPError as e:
            if e.response.status_code == 401:
                self.connection.authenticate()
            raise

    def info(self):
        url = "%s/%s" % (self.connection.auth.storage, self.container)
        r = self.connection.session.head(url, verify=True)
        r.raise_for_status()
        result = {
            'used': int(r.headers['X-Container-Bytes-Used']),
            'quota': int(r.headers['X-Container-Meta-Quota-Bytes']),
            'count': int(r.headers['X-Container-Object-Count']),
            'private': True if r.headers['X-Container-Meta-Type'] == 'private' else False
        }
        return result


class SelectelCloudObject(object):

    def __init__(self, path):
        self.path = path
        self.api_con = SelectelApi()
        self._content = None

    @property
    def content(self):
        if not self._content:
            iostring = StringIO()
            for chunk in self.api_con.get(self.path):
                iostring.write(chunk)
            iostring.seek(0)
            self._content = gzip.GzipFile(fileobj=iostring, mode='rb').read()
        return self._content

    @property
    def mimetype(self):
        return magic.from_buffer(self.content, mime=True)

    def read(self):
        return self.content

    def seek(self, *args, **kwargs):
        pass

    def tell(self):
        return len(self.content)

    def close(self):
        pass

    def delete(self):
        self.api_con.delete(self.path)
        self.path = None



# coding=utf-8
import hashlib
from datetime import datetime, timedelta
import requests


class Storage(object):
    SUPPORTED_ARCHIVES = ("tar", "tar.gz", "tar.bz2")

    def update_expired(fn):
        def wrapper(*args, **kwargs):
            auth = args[0].auth
            if auth.expired():
                args[0].authenticate()
            return fn(*args, **kwargs)
        return wrapper

    class Auth(object):
        THRESHOLD = 3600

        def __init__(self, token, storage, expires):
            self.token = token
            self.storage = storage[:-1]
            self.expires = datetime.now() + timedelta(seconds=int(expires))

        def expired(self):
            left = self.expires - datetime.now()
            return (left.total_seconds() < self.THRESHOLD)

    def __init__(self, user, key, timeout=3):
        self.url = "https://auth.selcdn.ru/"
        self.user = user
        self.timeout = timeout
        self.key = key
        self.authenticate()

    def authenticate(self):
        headers = {"X-Auth-User": self.user, "X-Auth-Key": self.key}
        r = requests.get(self.url, headers=headers, verify=True, timeout=self.timeout)
        if r.status_code != 204:
            raise Exception("Selectel: Unexpected status code: %s" %
                            r.status_code)
        auth = self.Auth(r.headers["X-Auth-Token"],
                         r.headers["X-Storage-Url"],
                         r.headers["X-Expire-Auth-Token"])
        self.auth = auth
        self.session = requests.Session()
        self.session.headers.update({"X-Auth-Token": self.auth.token})

    @update_expired
    def list(self, container, path=None):
        url = "%s/%s" % (self.auth.storage, container)
        params = {"format": "json"}
        if path is not None:
            if path.startswith("/"):
                path = path[1:]
            params["path"] = path
        r = self.session.get(url, params=params, verify=True, timeout=self.timeout)
        r.raise_for_status()

        def mapper(obj):
            dt = datetime.strptime(obj["last_modified"] + " GMT",
                                   "%Y-%m-%dT%H:%M:%S.%f %Z")
            result = {
                "content-type": obj["content_type"],
                "content-length": obj["bytes"],
                "hash": obj["hash"],
                "last-modified": dt
            }
            return result

        return {"/" + x["name"]: mapper(x) for x in r.json()}

    @update_expired
    def get(self, container, path, headers=None):
        url = "%s/%s%s" % (self.auth.storage, container, path)
        if headers is None:
            headers = {}
        r = self.session.get(url, headers=headers, verify=True, timeout=self.timeout)
        r.raise_for_status()
        return r.content

    @update_expired
    def get_stream(self, container, path, headers=None, chunk=2**20):
        url = "%s/%s%s" % (self.auth.storage, container, path)
        if headers is None:
            headers = {}
        r = self.session.get(url, headers=headers, stream=True, verify=True, timeout=self.timeout)
        r.raise_for_status()
        return r.iter_content(chunk_size=chunk)

    @update_expired
    def put(self, container, path, content, headers=None, extract=None):
        url = "%s/%s%s" % (self.auth.storage, container, path)
        if headers is None:
            headers = {}
        if extract in self.SUPPORTED_ARCHIVES:
            url += "?extract-archive=%s" % extract
            headers["Accept"] = "application/json"
        if not extract:
            headers["ETag"] = hashlib.md5(content).hexdigest()
        r = self.session.put(url, data=content, headers=headers, verify=True, timeout=self.timeout)
        r.raise_for_status()
        if extract in self.SUPPORTED_ARCHIVES:
            answer = r.json()
            return (answer["Number Files Created"], answer["Errors"])

    @update_expired
    def put_stream(self, container, path, descriptor,
                   headers=None, chunk=2**20, extract=None):
        url = "%s/%s%s" % (self.auth.storage, container, path)
        if headers is None:
            headers = {}
        if extract in self.SUPPORTED_ARCHIVES:
            url += "?extract-archive=%s" % extract
            headers["Accept"] = "application/json"

        def gen():
            data = descriptor.read(chunk)
            while data:
                yield data
                data = descriptor.read(chunk)

        r = self.session.put(url, data=gen(), headers=headers, verify=True, timeout=self.timeout)
        r.raise_for_status()
        if extract in self.SUPPORTED_ARCHIVES:
            answer = r.json()
            return (answer["Number Files Created"], answer["Errors"])

    @update_expired
    def put_file(self, container, path, filename, headers=None, extract=None):
        url = "%s/%s%s" % (self.auth.storage, container, path)
        if headers is None:
            headers = {}
        if extract in self.SUPPORTED_ARCHIVES:
            url += "?extract-archive=%s" % extract
            headers["Accept"] = "application/json"
        with open(filename, 'r+b') as file:
            r = self.session.put(url, data=file, headers=headers, verify=True, timeout=self.timeout)
            r.raise_for_status()
        if extract in self.SUPPORTED_ARCHIVES:
            answer = r.json()
            return (answer["Number Files Created"], answer["Errors"])

    @update_expired
    def remove(self, container, path, force=False):
        url = "%s/%s%s" % (self.auth.storage, container, path)
        r = self.session.delete(url, verify=True, timeout=self.timeout)
        if force:
            if r.status_code == 404:
                return r.headers
            else:
                r.raise_for_status()
        else:
            r.raise_for_status()

    @update_expired
    def copy(self, container, src, dst, headers=None):
        dst = "%s/%s%s" % (self.auth.storage, container, dst)
        src = "%s%s" % (container, src)
        if headers is None:
            headers = {}
        headers["X-Copy-From"] = src
        r = self.session.put(dst, headers=headers, verify=True, timeout=self.timeout)
        r.raise_for_status()

    @update_expired
    def info(self, container, path=None):
        if path is None:
            url = "%s/%s" % (self.auth.storage, container)
        else:
            url = "%s/%s%s" % (self.auth.storage, container, path)
        r = self.session.head(url, verify=True, timeout=self.timeout)
        r.raise_for_status()
        if path is None:
            result = {
                "count": int(r.headers["X-Container-Object-Count"]),
                "usage": int(r.headers["X-Container-Bytes-Used"]),
                "public": (r.headers.get("X-Container-Meta-Type") == "public"),
                "tx": int(r.headers["X-Transfered-Bytes"]),
                "rx": int(r.headers["X-Received-Bytes"])
            }
        else:
            dt = datetime.strptime(r.headers["Last-Modified"],
                                   "%a, %d %b %Y %H:%M:%S %Z")
            result = {
                "content-length": int(r.headers["Content-Length"]),
                "last-modified": dt,
                "hash": r.headers["ETag"],
                "content-type": r.headers["Content-Type"],
                "downloads": int(r.headers["X-Object-Downloads"])
            }
        return result

    @update_expired
    def create(self, container, public=False, headers=None):
        url = "%s/%s" % (self.auth.storage, container)
        if headers is None:
            headers = {}
        if public:
            headers["X-Container-Meta-Type"] = "public"
        else:
            headers["X-Container-Meta-Type"] = "private"
        r = self.session.put(url, headers=headers, verify=True, timeout=self.timeout)
        r.raise_for_status()

    @update_expired
    def drop(self, container, force=False, recursive=False):
        url = "%s/%s" % (self.auth.storage, container)
        if recursive:
            for filename in self.list(container):
                self.remove(container, filename, force=force)
        r = self.session.delete(url, verify=True, timeout=self.timeout)
        if force:
            if r.status_code == 404:
                pass
            else:
                r.raise_for_status()
        else:
            r.raise_for_status()


class Container(object):
    METHODS = ["list", "get", "get_stream", "put",
               "put_stream", "put_file", "remove",
               "copy", "info"]

    def __init__(self, auth, key, name):
        self.name = name
        self.storage = Storage(auth, key)

        def make_method(name):
            def method(*args, **kwargs):
                fn = getattr(self.storage, name)
                return fn(self.name, *args, **kwargs)
            return method

        for name in self.METHODS:
            setattr(self, name, make_method(name))



#! coding: utf-8
from __future__ import absolute_import, unicode_literals

import gzip
import os
import types
from hashlib import md5
from StringIO import StringIO

from mongoengine.fields import BaseField
from selectel_storage.base import SelectelCloudObject, SelectelApi


class SelectelStorageField(BaseField):

    def __init__(self, *args, **kwargs):
        self.root = kwargs.pop('root', "/")
        if not self.root.startswith('/'):
            self.root = "/" + self.root
        super(SelectelStorageField, self).__init__(*args, **kwargs)

    def get_path(self, content):
        filename = md5(content).hexdigest()
        return self.root + os.path.join(filename[:2], filename[2:4], filename + ".gz")

    def to_python(self, value):
        if isinstance(value, types.StringTypes):
            value = SelectelCloudObject(value)
        return value

    def to_mongo(self, value):
            try:
                if hasattr(value, 'read'):
                    if hasattr(value, 'seek'):
                        value.seek(0)
                    content = value.read()
                    if hasattr(value, 'seek'):
                        value.seek(0)

                    path = self.get_path(content)
                    compers_obj = StringIO()
                    compers_obj_gzip = gzip.GzipFile(fileobj=compers_obj, mode='wb')
                    compers_obj_gzip.write(content)
                    compers_obj_gzip.close()
                    self.api_conn.add(path, compers_obj.getvalue())
                    return path
            except Exception, e:
                self.error(unicode(e))

    def prepare_query_value(self, op, value):
        if value:
            return value.path

    def validate(self, value):
        if not hasattr(value, 'read'):
            self.error("type '{}' it does not implement the method 'read'".format(type(value)))

    @property
    def api_conn(self):
        return SelectelApi()



from setuptools import setup, find_packages


VERSION = "0.0.21"

setup(
    name='selectel-storage',
    description="",
    version=VERSION,
    url='https://github.com/KokocGroup/selectel_storage',
    download_url='https://github.com/KokocGroup/selectel_storage/tarball/v{}'.format(VERSION),
    packages=find_packages(),
    install_requires=[
        'mongoengine==0.10.0',
        'python-magic==0.4.10',
        'requests==2.10.0'
    ],
)
